#!/usr/bin/env python3
"""
"""
import hashlib
import os
import pytest

from autouri.httpurl import HTTPURL
from autouri.abspath import AbsPath
from autouri.gcsuri import GCSURI
from autouri.s3uri import S3URI

from .files import (
    v6_txt_contents,
    j1_json,
    v41_json,
    v421_tsv,
    v5_csv,
    v6_txt,
)


GCS_ENDPOINT_URL = 'https://storage.googleapis.com/'


def pytest_addoption(parser):
    parser.addoption(
        '--ci-prefix', required=True,
        help='Prefix for CI test.'
    )
    parser.addoption(
        '--s3-root', default='s3://encode-test-autouri/tmp',
        help='S3 root path for CI test. '
             'This S3 bucket must be configured without versioning. '
             'Make it publicly accessible. '
             'Read access for everyone is enough for testing. '
    )
    parser.addoption(
        '--s3-public-url-test-v6-file', default='s3://encode-test-autouri/tmp/v6.txt',
        help='Write "v6: Hello World" to this file named "v6.txt" '
             'and grant "Read object" permission on it. '
             'Since S3 object does not inherit ACL from bucket/parent '
             'and S3URI does not have methods to control ACL of an object '
             'so this is the only way to test get_public_url(self) method in '
             'S3URI.'
    )
    parser.addoption(
        '--gcs-root', default='gs://encode-test-autouri/tmp',
        help='GCS root path for CI test. '
             'This GCS bucket must be publicly accessible '
             '(read access for everyone is enough for testing).'
    )
    parser.addoption(
        '--gcs-root-url', default='gs://encode-test-autouri/tmp_url',
        help='GCS root path for CI test for URLs. '
             'This GCS bucket must be public so that '
             'anyone can access to it via an URL with '
             'an endpoint GCS_ENDPOINT_URL. '
    )
    parser.addoption(
        '--gcp-private-key-file', required=True,
        help='GCP private key file (JSON format) to test presigning GCS URIs. '
             'Generate one for a service account that has an admin access to both '
             '--gcs-root and --gcs-root-url.'
    )


@pytest.fixture(scope="session")
def ci_prefix(request):
    return request.config.getoption("--ci-prefix").rstrip('/')


@pytest.fixture(scope="session")
def s3_root(request):
    """S3 root to generate test S3 URIs on.
    """
    return request.config.getoption("--s3-root").rstrip('/')


@pytest.fixture(scope="session")
def s3_public_url_test_v6_file(request):
    return request.config.getoption("--s3-public-url-test-v6-file")


@pytest.fixture(scope="session")
def gcs_root(request):
    """GCS root to generate test GCS URIs on.
    """
    return request.config.getoption("--gcs-root").rstrip('/')


@pytest.fixture(scope="session")
def gcs_root_url(request):
    """GCS root to generate test URLs on. This GCS bucket should be public.
    """
    return request.config.getoption("--gcs-root-url").rstrip('/')


@pytest.fixture(scope="session")
def gcp_private_key_file(request):
    """GCS private key file to test presigning GCS URIs.
    """
    return request.config.getoption("--gcp-private-key-file")


@pytest.fixture(scope='session')
def local_test_path(tmpdir_factory, ci_prefix):
    return tmpdir_factory.mktemp(ci_prefix).realpath()


@pytest.fixture(scope='session')
def s3_test_path(s3_root, ci_prefix):
    return '{s3_root}/{ci_prefix}'.format(
        s3_root=s3_root, ci_prefix=ci_prefix)


@pytest.fixture(scope='session')
def gcs_test_path(gcs_root, ci_prefix):
    return '{gcs_root}/{ci_prefix}'.format(
        gcs_root=gcs_root, ci_prefix=ci_prefix)


@pytest.fixture(scope='session')
def gcs_test_path_url(gcs_root_url, ci_prefix):
    return '{gcs_root_url}/{ci_prefix}'.format(
        gcs_root_url=gcs_root_url, ci_prefix=ci_prefix)


@pytest.fixture(scope='session')
def gcs_test_path_self_ref(gcs_test_path):
    return '{gcs_test_path}/self_ref'.format(
        gcs_test_path=gcs_test_path)


@pytest.fixture(scope='session')
def url_test_path(gcs_root_url, ci_prefix):
    url_root = gcs_root_url.replace('gs://', GCS_ENDPOINT_URL, 1)
    return '{url_root}/{ci_prefix}'.format(
        url_root=url_root, ci_prefix=ci_prefix)


@pytest.fixture(scope='session')
def mixed_local_test_path(local_test_path):
    d = os.path.join(local_test_path, 'mixed')
    os.makedirs(d, exist_ok=True)
    return d


@pytest.fixture(scope='session')
def mixed_s3_test_path(s3_test_path):
    return '{s3_test_path}/mixed'.format(
        s3_test_path=s3_test_path)


@pytest.fixture(scope='session')
def mixed_gcs_test_path(gcs_test_path):
    return '{gcs_test_path}/mixed'.format(
        gcs_test_path=gcs_test_path)


@pytest.fixture(scope='session')
def mixed_gcs_test_path_url(gcs_test_path_url):
    return '{gcs_test_path_url}/mixed'.format(
        gcs_test_path_url=gcs_test_path_url)


@pytest.fixture(scope='session')
def mixed_url_test_path(url_test_path):
    return '{url_test_path}/mixed'.format(
        url_test_path=url_test_path)


@pytest.fixture(scope="session")
def local_j1_json(local_test_path):
    return j1_json(local_test_path, make=True)


@pytest.fixture(scope="session")
def local_v41_json(local_test_path):
    return v41_json(local_test_path, make=True)


@pytest.fixture(scope="session")
def local_v421_tsv(local_test_path):
    return v421_tsv(local_test_path, make=True)


@pytest.fixture(scope="session")
def local_v5_csv(local_test_path):
    return v5_csv(local_test_path, make=True)


@pytest.fixture(scope="session")
def local_v6_txt(local_test_path):
    return v6_txt(local_test_path, make=True)


@pytest.fixture(scope="session")
def s3_j1_json(s3_test_path):
    return j1_json(s3_test_path, make=True)


@pytest.fixture(scope="session")
def s3_v41_json(s3_test_path):
    return v41_json(s3_test_path, make=True)


@pytest.fixture(scope="session")
def s3_v421_tsv(s3_test_path):
    return v421_tsv(s3_test_path, make=True)


@pytest.fixture(scope="session")
def s3_v5_csv(s3_test_path):
    return v5_csv(s3_test_path, make=True)


@pytest.fixture(scope="session")
def s3_v6_txt(s3_test_path):
    return v6_txt(s3_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_j1_json(gcs_test_path):
    return j1_json(gcs_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_v41_json(gcs_test_path):
    return v41_json(gcs_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_v421_tsv(gcs_test_path):
    return v421_tsv(gcs_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_v5_csv(gcs_test_path):
    return v5_csv(gcs_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_v6_txt(gcs_test_path):
    return v6_txt(gcs_test_path, make=True)


@pytest.fixture(scope="session")
def gcs_j1_json_url(gcs_test_path_url, url_test_path):
    return j1_json(gcs_test_path_url, make=True,
        prefix_v41=url_test_path,
        prefix_v421=url_test_path,
        prefix_v5=url_test_path)


@pytest.fixture(scope="session")
def gcs_v41_json_url(gcs_test_path_url):
    return v41_json(gcs_test_path_url, make=True)


@pytest.fixture(scope="session")
def gcs_v421_tsv_url(gcs_test_path_url, url_test_path):
    return v421_tsv(gcs_test_path_url, make=True,
        prefix_v5=url_test_path,
        prefix_v1=url_test_path)


@pytest.fixture(scope="session")
def gcs_v5_csv_url(gcs_test_path_url, url_test_path):
    return v5_csv(gcs_test_path_url, make=True,
        prefix_v6=url_test_path)


@pytest.fixture(scope="session")
def gcs_v6_txt_url(gcs_test_path_url):
    return v6_txt(gcs_test_path_url, make=True)


@pytest.fixture(scope="session")
def gcs_j1_json_self_ref(gcs_test_path_self_ref):
    return j1_json(gcs_test_path_self_ref, make=True)


@pytest.fixture(scope="session")
def gcs_v41_json_self_ref(gcs_test_path_self_ref):
    return v41_json(gcs_test_path_self_ref, make=True)


@pytest.fixture(scope="session")
def gcs_v421_tsv_self_ref(gcs_test_path_self_ref):
    return v421_tsv(gcs_test_path_self_ref, make=True, make_link_to_j1_json=True)


@pytest.fixture(scope="session")
def gcs_v5_csv_self_ref(gcs_test_path_self_ref):
    return v5_csv(gcs_test_path_self_ref, make=True)


@pytest.fixture(scope="session")
def gcs_v6_txt_self_ref(gcs_test_path_self_ref):
    return v6_txt(gcs_test_path_self_ref, make=True)


@pytest.fixture(scope="session")
def url_j1_json(gcs_j1_json_url, url_test_path):
    """URL is read-only. So this is a link to the actual file on GCS.
    """
    return j1_json(url_test_path, make=False)


@pytest.fixture(scope="session")
def url_v41_json(gcs_v41_json_url, url_test_path):
    """URL is read-only. So this is a link to the actual file on GCS.
    """
    return v41_json(url_test_path, make=False)


@pytest.fixture(scope="session")
def url_v421_tsv(gcs_v421_tsv_url, url_test_path):
    """URL is read-only. So this is a link to the actual file on GCS.
    """
    return v421_tsv(url_test_path, make=False)


@pytest.fixture(scope="session")
def url_v5_csv(gcs_v5_csv_url, url_test_path):
    """URL is read-only. So this is a link to the actual file on GCS.
    """
    return v5_csv(url_test_path, make=False)


@pytest.fixture(scope="session")
def url_v6_txt(gcs_v6_txt_url, url_test_path):
    """URL is read-only. So this is a link to the actual file on GCS.
    """
    return v6_txt(url_test_path, make=False)


@pytest.fixture(scope="session")
def mixed_j1_json(
    mixed_local_test_path,
    mixed_s3_test_path,
    mixed_gcs_test_path,
    mixed_url_test_path):
    return j1_json(mixed_gcs_test_path, make=True,
        prefix_v41=mixed_url_test_path,
        prefix_v421=mixed_s3_test_path,
        prefix_v5=mixed_local_test_path)


@pytest.fixture(scope="session")
def mixed_v41_json(mixed_gcs_test_path_url, mixed_url_test_path):
    """Actual file is created on GCS then it will have a correspondig URL.
    """
    gcs_path = v41_json(mixed_gcs_test_path_url, make=True)
    url = gcs_path.replace(mixed_gcs_test_path_url, mixed_url_test_path, 1)
    return url


@pytest.fixture(scope="session")
def mixed_v421_tsv(
    mixed_s3_test_path,
    mixed_gcs_test_path,
    mixed_local_test_path):
    return v421_tsv(mixed_s3_test_path, make=True,
        prefix_v5=mixed_local_test_path,
        prefix_v1=mixed_gcs_test_path)


@pytest.fixture(scope="session")
def mixed_v5_csv(
    mixed_local_test_path,
    mixed_s3_test_path):
    return v5_csv(mixed_local_test_path, make=True,
        prefix_v6=mixed_s3_test_path)


@pytest.fixture(scope="session")
def mixed_v6_txt(mixed_s3_test_path):
    return v6_txt(mixed_s3_test_path, make=True)


@pytest.fixture(scope="session")
def v6_txt_size():
    return len(v6_txt_contents())


@pytest.fixture(scope="session")
def v6_txt_md5_hash():
    return hashlib.md5(v6_txt_contents().encode()).hexdigest()



#!/usr/bin/env python3
"""Files for testing

Use v6_txt to test basic/simple methods of AutoURI.

Use j1_json to test recursive localization of a file containing file paths in it.
j1_json has TSV/CSV files, which can also be recursively localized.

See j1_json_contents() for its detailed structure.
    j1_json
        v41_json
        v421_tsv
            v5_csv
                v6_txt
            v6_txt
            v1_json (optionally to check self-reference)
        v5_csv
            v6_txt
"""
from autouri.autouri import AutoURI
from autouri.loc_aux import recurse_json, recurse_tsv, recurse_csv


def j1_json_contents(prefix, prefix_v41=None, prefix_v421=None, prefix_v5=None, loc_suffix=''):
    if prefix_v41 is None:
        prefix_v41 = prefix
    if prefix_v421 is None:
        prefix_v421 = prefix
    if prefix_v5 is None:
        prefix_v5 = prefix
    return '''{{
        "k1": "v1",
        "k2": 1,
        "k3": 2.2,
        "k4": {{
            "k41": "{prefix_v41}/v41.json",
            "k42": {{
                "k421": "{prefix_v421}/deeper/v421{loc_suffix}.tsv"
            }},
            "k43": null
        }},
        "k5": "{prefix_v5}/even/deeper/v5{loc_suffix}.csv"
    }}
    '''.format(        
        prefix_v41=prefix_v41,
        prefix_v421=prefix_v421,
        prefix_v5=prefix_v5,
        loc_suffix=loc_suffix)


def v41_json_contents():
    """JSON without any files defined in it.
    """
    return '{ "k1": "v1" }'


def v421_tsv_contents(prefix, prefix_v5=None, prefix_v1=None, prefix_v6=None, loc_suffix='', make_link_to_j1_json=False):
    """
    Args:
        make_link_to_j1_json:
            Link j1.json in this TSV file to to intentionally make a self reference
            , and hence an infinite loop while recursive localization
            j1.json -> v421.tsv -> j1.json -> ...
    """
    if prefix_v5 is None:
        prefix_v5 = prefix
    if prefix_v1 is None:
        prefix_v1 = prefix
    if prefix_v6 is None:
        prefix_v6 = prefix
    arr = [
        'k1\tv1',
        'k2\t1',
        'k3\t2.2',
        'k4\tnot/absolute/path',
        'k5\t{prefix_v5}/even/deeper/v5{loc_suffix}.csv',
        'k6\t{prefix_v6}/v6.txt'
    ]
    if make_link_to_j1_json:
        arr.append('k7\t{prefix_v1}/j1.json')
    s = '\n'.join(arr)
    return s.format(
        prefix_v5=prefix_v5,
        prefix_v1=prefix_v1,
        prefix_v6=prefix_v6,
        loc_suffix=loc_suffix)


def v5_csv_contents(prefix, prefix_v6=None):
    if prefix_v6 is None:
        prefix_v6 = prefix
    s = '\n'.join([
        'k1,v1',
        'k2,1',
        'k3,2.2',
        'k4,not/absolute/path',
        'k5,s33://not-valid-bucket-address',
        'k6,{prefix_v6}/v6.txt'
    ])
    return s.format(prefix_v6=prefix_v6)


def v6_txt_contents():
    return 'v6: Hello World'


def j1_json(prefix, prefix_v41=None, prefix_v421=None, prefix_v5=None, loc_suffix='', make=False):
    u = '{prefix}/j1{loc_suffix}.json'.format(
        prefix=prefix, loc_suffix=loc_suffix)
    if make:
        AutoURI(u).write(j1_json_contents(
            prefix=prefix,
            prefix_v41=prefix_v41,
            prefix_v421=prefix_v421,
            prefix_v5=prefix_v5,
            loc_suffix=loc_suffix))
    return u


def v41_json(prefix, make=False):
    u = '{prefix}/v41.json'.format(
        prefix=prefix)
    if make:
        AutoURI(u).write(v41_json_contents())
    return u


def v421_tsv(prefix, prefix_v5=None, prefix_v1=None, prefix_v6=None, loc_suffix='', make=False, make_link_to_j1_json=False):
    u = '{prefix}/deeper/v421{loc_suffix}.tsv'.format(
        prefix=prefix, loc_suffix=loc_suffix)
    if make:
        AutoURI(u).write(v421_tsv_contents(
            prefix=prefix,
            prefix_v5=prefix_v5,
            prefix_v1=prefix_v1,
            prefix_v6=prefix_v6,
            loc_suffix=loc_suffix,
            make_link_to_j1_json=make_link_to_j1_json))
    return u


def v5_csv(prefix, prefix_v6=None, make=False):
    u = '{prefix}/even/deeper/v5.csv'.format(
        prefix=prefix)
    if make:
        AutoURI(u).write(v5_csv_contents(
            prefix=prefix,
            prefix_v6=prefix_v6))
    return u


def v6_txt(prefix, make=False):
    u = '{prefix}/v6.txt'.format(
        prefix=prefix)
    if make:
        AutoURI(u).write(v6_txt_contents())
    return u


def common_paths():
    return [
        '/testing/abspath',
        '/testing/abspath/',
        '~/os/expandable',
        '~~/os/expandable',
        '~/~/os/expandable',
        'test/ok/man.csv',
        'http://hello.world.com/ok.txt',
        'https://hello.world.com/ok.txt',
        'http://hello.world.com/ok.txt?parameter1=true&parameter2=false',
        'https://hello.world.com/ok.txt?parameter1=true&parameter2=false',
        'http:/hello.world.com/notok.txt',
        'ftp:/hello.world.com/notok.txt',
        'dx://dnanexus-prj/not/supported.txt',
        'file:/notok.txt',
        'file://hostname/notok.txt',
        's3://hello-world/ok.txt',
        's3:/hello-world/not-ok.txt',
        's3:\\hello-world\\not-ok.txt',
        'gs://hello-world/ok.txt',
        'gs:/hello-world/not-ok.txt',
        'gs:\\hello-world\\not-ok.txt',
        '!@#:;$@!#$F',
    ]


def recurse_raise_if_uri_not_exist(uri):
    uri = AutoURI(uri)
    if uri.is_valid:
        if uri.exists:
            if uri.ext == '.json':
                recurse_json(uri.read(), recurse_raise_if_uri_not_exist)
            elif uri.ext == '.tsv':
                recurse_tsv(uri.read(), recurse_raise_if_uri_not_exist)
            elif uri.ext == '.csv':
                recurse_csv(uri.read(), recurse_raise_if_uri_not_exist)
        else:
            raise Exception('URI is a valid path but does not exist.')
    return None, False



#!/usr/bin/env python3
"""Basic recursive localization testing is already done each class' testing module.
Here, we test more extreme/edge cases for recursive localization:
    1) Direct/indirect self reference
    2) Remote files with mixed storage types
"""
import os
import pytest
import time
from typing import Any, Tuple, Union

from autouri.abspath import AbsPath
from autouri.autouri import AutoURI, URIBase, AutoURIRecursionError
from autouri.httpurl import ReadOnlyStorageError

from .files import (
    v6_txt_contents,
    common_paths,
    recurse_raise_if_uri_not_exist)


def test_localize_self_ref(
        local_test_path,
        gcs_j1_json_self_ref,
        gcs_v41_json_self_ref,
        gcs_v421_tsv_self_ref,
        gcs_v5_csv_self_ref,
        gcs_v6_txt_self_ref,
    ) -> Tuple[str, bool]:
    """Test detection of direct/indirect self reference
    while localizing GCS files on local storage.
    This is not only for detecting self references but also for any deep recursion
    beyond the depth limit (10 by default)

    Indirect self referencing in GCS files suffixed with _self_ref:
        v1.json -> v421.tsv -> v1.json -> ...
    """
    loc_prefix = os.path.join(local_test_path, 'test_localize_self_ref')

    # localization from remote storages
    for j1_json in (gcs_j1_json_self_ref,):
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()

        with pytest.raises(AutoURIRecursionError):
            loc_uri, localized = AbsPath.localize(
                u_j1_json,
                recursive=True,
                loc_prefix=loc_prefix_)


def test_localize_mixed(
        local_test_path,
        mixed_j1_json,
        mixed_v41_json,
        mixed_v421_tsv,
        mixed_v5_csv,
        mixed_v6_txt,
    ) -> Tuple[str, bool]:
    """Test recursive localization of files on mixed storages
    Target is local storage.
    """
    loc_prefix = os.path.join(local_test_path, 'test_localize_mixed')

    # localization from remote storages
    for j1_json in (mixed_j1_json,):
        u_j1_json = AutoURI(j1_json)
        loc_prefix_ = loc_prefix + u_j1_json.__class__.get_loc_suffix()

        loc_uri, localized = AbsPath.localize(
            u_j1_json,
            recursive=True,
            return_flag=True,
            loc_prefix=loc_prefix_)
        # check if all URIs defeind in localized JSON file exist
        recurse_raise_if_uri_not_exist(loc_uri)



"""GCS Bucket configuration:
    Object versioning must be turned off
        Check it with "gsutil versioning get gs://BUCKET-NAME"
        https://cloud.google.com/storage/docs/object-versioning
"""
import logging
import os
import requests
import time
from datetime import timedelta
from filelock import BaseFileLock
from google.api_core.exceptions import NotFound, Forbidden, GatewayTimeout, ServiceUnavailable
from google.cloud import storage
from google.cloud.storage import Blob
from google.oauth2.service_account import Credentials
from subprocess import check_call, PIPE, CalledProcessError
from tempfile import NamedTemporaryFile, TemporaryDirectory
from typing import Tuple, Optional
from urllib3.exceptions import HTTPError
from .autouri import URIBase, AutoURI, logger
from .metadata import URIMetadata, get_seconds_from_epoch, parse_md5_str


logger = logging.getLogger(__name__)


class GCSURILock(BaseFileLock):
    """Slow but stable locking with using GCS temporary_hold
    """
    def __init__(
        self, lock_file, thread_id=-1, timeout=900, poll_interval=10.0, no_lock=False):
        super().__init__(lock_file, timeout=timeout)
        self._poll_interval = poll_interval
        self._thread_id = thread_id

    def acquire(self, timeout=None, poll_intervall=5.0):
        """Use self._poll_interval instead of poll_intervall in args
        """
        super().acquire(timeout=timeout, poll_intervall=self._poll_interval)

    def _acquire(self):
        u = GCSURI(self._lock_file, thread_id=self._thread_id)
        blob, bucket_obj = u.get_blob(new=True)
        if blob is not None:
            try:
                blob.upload_from_string('')
                blob.temporary_hold = True
                blob.patch()
                self._lock_file_fd = id(self)
            except (Forbidden, GatewayTimeout, NotFound, ServiceUnavailable):
                pass
        return None

    def _release(self):
        u = GCSURI(self._lock_file, thread_id=self._thread_id)
        blob, _ = u.get_blob()
        if blob is not None:
            blob.temporary_hold = False
            try:
                blob.patch()
                blob.delete()
                self._lock_file_fd = None
            except (NotFound,):
                pass
        return None


class GCSURI(URIBase):
    """
    Class constants:
        LOC_PREFIX (inherited):
            Path prefix for localization. Inherited from URIBase class.
        PRIVATE_KEY_FILE:
            Path for private key file used to get presigned URLs
        DURATION_PRESIGNED_URL:
            Duration for presigned URLs in seconds.
        RETRY_BUCKET:
            Number of retrial to access a bucket.
        RETRY_BUCKET_DELAY:
            Delay for each retrial in seconds.
        USE_GSUTIL_FOR_S3 (experimental):
            This is only for direct transfer between S3 and GCS buckets.
            WARNING:
                gsutil must be configured correctly to have all
                AWS credentials in ~/.boto file.
                Run "aws configure" first and then
                run "gsutil config" to generate corrensponding ~/.boto file.

    Protected class constants:
        _CACHED_GCS_CLIENT_PER_THREAD:
            Per-thread GCS client object is required since 
            GCS client is not thread-safe.
        _CACHED_PRESIGNED_URLS:
            Can use cached presigned URLs.
        _GCS_PUBLIC_URL_FORMAT:
            End point for a bucket with public access + key path
    """
    PRIVATE_KEY_FILE: str = ''
    DURATION_PRESIGNED_URL: int = 4233600

    RETRY_BUCKET: int = 3
    RETRY_BUCKET_DELAY: int = 1
    USE_GSUTIL_FOR_S3: bool = False

    _CACHED_GCS_CLIENT_PER_THREAD = {}
    _CACHED_PRESIGNED_URLS = {}
    _GCS_PUBLIC_URL_FORMAT = 'http://storage.googleapis.com/{bucket}/{path}'

    _LOC_SUFFIX = '.gcs'
    _SCHEMES = ('gs://',)

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)

    def _get_lock(self, timeout=None, poll_interval=None):
        if timeout is None:
            timeout = GCSURI.LOCK_TIMEOUT
        if poll_interval is None:
            poll_interval = GCSURI.LOCK_POLL_INTERVAL
        return GCSURILock(
            self._uri + GCSURI.LOCK_FILE_EXT,
            thread_id=self._thread_id,
            timeout=timeout,
            poll_interval=poll_interval)

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        ex, mt, sz, md5 = False, None, None, None

        try:
            b, _ = self.get_blob()
            if b is not None:
                # make keys lower-case
                h = {k.lower(): v for k, v in b._properties.items()}
                ex = True

                if not skip_md5:
                    if 'md5hash' in h:
                        md5 = parse_md5_str(h['md5hash'])
                    elif 'etag' in h:
                        md5 = parse_md5_str(h['etag'])
                    if md5 is None:
                        # make_md5_file is ignored for GCSURI
                        md5 = self.md5_from_file

                if 'size' in h:
                    sz = int(h['size'])

                if 'updated' in h:
                    mt = get_seconds_from_epoch(h['updated'])
                elif 'timecreated' in h:
                    mt = get_seconds_from_epoch(h['timecreated'])

        except Exception as e:
            pass

        return URIMetadata(
            exists=ex,
            mtime=mt,
            size=sz,
            md5=md5)

    def read(self, byte=False):
        blob, _ = self.get_blob()
        b = blob.download_as_string()
        if byte:
            return b
        return b.decode()

    def _write(self, s):
        blob, _ = self.get_blob(new=True)
        blob.upload_from_string(s)
        # blob.update()
        return

    def _rm(self):
        blob, _ = self.get_blob()
        blob.delete()
        return

    def _cp(self, dest_uri):
        """Copy from GCSURI to 
            GCSURI
            S3URI: can use gsutil for direct transfer if USE_GSUTIL_FOR_S3 == True
            AbsPath
        """
        from .s3uri import S3URI
        from .abspath import AbsPath

        dest_uri = AutoURI(dest_uri)

        if isinstance(dest_uri, (GCSURI, AbsPath)):            
            src_blob, src_bucket = self.get_blob()

            if src_blob is None:
                raise ValueError('Blob does not exist for {f}'.format(f=self._uri))

            if isinstance(dest_uri, GCSURI):
                _, dest_path = dest_uri.get_bucket_path()
                _, dest_bucket = dest_uri.get_blob()
                src_bucket.copy_blob(src_blob, dest_bucket, dest_path)
                return True

            elif isinstance(dest_uri, AbsPath):
                dest_uri.mkdir_dirname()
                # mtime is not updated without update().
                src_blob.update()                
                src_blob.download_to_filename(dest_uri._uri)
                return True

        elif isinstance(dest_uri, S3URI):
            if GCSURI.USE_GSUTIL_FOR_S3:
                rc = check_call(['gsutil', '-q', 'cp', self._uri, dest_uri._uri])
                return rc == 0
            else:
                # use local temporary file instead
                with TemporaryDirectory() as tmp_d:
                    dest_uri_local = AbsPath(os.path.join(tmp_d, self.basename))
                    # lockless copy
                    self.cp(dest_uri=dest_uri_local, no_lock=True, no_checksum=True)
                    dest_uri_local.cp(dest_uri=dest_uri, no_lock=True, no_checksum=True)
                return True

        return False

    def _cp_from(self, src_uri):
        """Copy to GCSURI from
            S3URI: can use gsutil for direct transfer if USE_GSUTIL_FOR_S3 == True
            AbsPath
            HTTPURL
        """
        from .s3uri import S3URI
        from .abspath import AbsPath
        from .httpurl import HTTPURL

        src_uri = AutoURI(src_uri)

        if isinstance(src_uri, AbsPath):
            blob, _ = self.get_blob(new=True)
            blob.upload_from_filename(src_uri._uri)
            return True

        elif isinstance(src_uri, S3URI):
            if GCSURI.USE_GSUTIL_FOR_S3:
                rc = check_call(['gsutil', '-q', 'cp', src_uri._uri, self._uri])
                return rc == 0
            else:
                # use local temporary file instead
                with TemporaryDirectory() as tmp_d:
                    dest_uri_local = AbsPath(os.path.join(tmp_d, self.basename))
                    src_uri.cp(dest_uri=dest_uri_local, no_lock=True, no_checksum=True)
                    dest_uri_local.cp(dest_uri=self, no_lock=True, no_checksum=True)
                return True

        elif isinstance(src_uri, HTTPURL):
            r = requests.get(
                src_uri._uri, stream=True, allow_redirects=True,
                headers=requests.utils.default_headers())
            r.raise_for_status()
            with NamedTemporaryFile() as fp:
                for chunk in r.iter_content(HTTPURL.get_http_chunk_size()):
                    fp.write(chunk)
                fp.seek(0)
                blob, _ = self.get_blob(new=True)
                blob.upload_from_file(fp)
            return True
        return False

    def get_blob(self, new=False) -> Blob:
        """GCS client() has a bug that shows an outdated version of a file
        when using Blob() without update().
        For read-only functions (e.g. read()), need to directly call
        cl.get_bucket(bucket).get_blob(path) instead of using Blob() class.

        Also, GCS client() is not thread-safe and it fails for a variety of reasons.
        Retry several times for whatever reasons.

        Returns:
            blob: Blob object or None
            bucket_obj: Bucket object or None
        """
        bucket, path = self.get_bucket_path()
        cl = GCSURI.get_gcs_client(self._thread_id)

        bucket_obj = None
        blob = None
        for retry in range(GCSURI.RETRY_BUCKET):
            try:
                bucket_obj = cl.get_bucket(bucket)
                blob = bucket_obj.get_blob(path)
                if new and blob is None:
                    blob = Blob(name=path, bucket=bucket_obj)
                break
            except NotFound:
                raise
            except PermissionDenied:
                raise
            except:
                time.sleep(GCSURI.RETRY_BUCKET_DELAY)
        return blob, bucket_obj

    def get_bucket_path(self) -> Tuple[str, str]:
        """Returns a tuple of URI's S3 bucket and path.
        """
        bucket, path = self.uri_wo_scheme.split(GCSURI.get_path_sep(), 1)
        return bucket, path        

    def get_presigned_url(self, duration=None, private_key_file=None, use_cached=False) -> str:
        """
        Args:
            duration: Duration in seconds. This is ignored if use_cached is on.
            use_cached: Use a cached URL. 
        """
        cache = GCSURI._CACHED_PRESIGNED_URLS
        if use_cached:
            if cache is not None and self._uri in cache:
                return cache[self._uri]
        # if not self.exists:
        #     raise Exception('File does not exist. f={f}'.format(self._uri))
        if private_key_file is None:            
            private_key_file = os.path.expanduser(GCSURI.PRIVATE_KEY_FILE)
        else:
            private_key_file = os.path.expanduser(private_key_file)
        if not os.path.exists(private_key_file):
            raise Exception('GCS private key file not found. f:{f}'.format(
                f=private_key_file))
        credentials = Credentials.from_service_account_file(private_key_file)
        duration = duration if duration is not None else GCSURI.DURATION_PRESIGNED_URL        
        blob, _ = self.get_blob()
        if blob is None:
            raise ValueError('Blob does not exist for {f}'.format(f=self._uri))
        url = blob.generate_signed_url(
            expiration=timedelta(seconds=duration),
            credentials=credentials)
        cache[self._uri] = url
        return url

    def get_public_url(self) -> str:
        bucket, path = self.get_bucket_path()
        return GCSURI._GCS_PUBLIC_URL_FORMAT.format(bucket=bucket, path=path)

    @staticmethod
    def get_gcs_client(thread_id) -> storage.Client:
        if thread_id in GCSURI._CACHED_GCS_CLIENT_PER_THREAD:
            return GCSURI._CACHED_GCS_CLIENT_PER_THREAD[thread_id]
        else:
            cl = storage.Client()
            GCSURI._CACHED_GCS_CLIENT_PER_THREAD[thread_id] = cl
            return cl

    @staticmethod
    def init_gcsuri(
        loc_prefix: Optional[str]=None,
        private_key_file: Optional[str]=None,
        duration_presigned_url: Optional[int]=None,
        retry_bucket: Optional[int]=None,
        retry_bucket_delay: Optional[int]=None,
        use_gsutil_for_s3: Optional[bool]=None):
        if loc_prefix is not None:
            GCSURI.LOC_PREFIX = loc_prefix
        if private_key_file is not None:
            GCSURI.PRIVATE_KEY_FILE = private_key_file
        if duration_presigned_url is not None:
            GCSURI.DURATION_PRESIGNED_URL = duration_presigned_url
        if retry_bucket is not None:
            GCSURI.RETRY_BUCKET = retry_bucket
        if retry_bucket_delay is not None:
            GCSURI.RETRY_BUCKET_DELAY = retry_bucket_delay
        if use_gsutil_for_s3 is not None:
            GCSURI.USE_GSUTIL_FOR_S3 = use_gsutil_for_s3



#!/usr/bin/env python3
import pytest
import time
from binascii import hexlify

from autouri.metadata import (
    URIMetadata,
    get_seconds_from_epoch,
    base64_to_hex,
    parse_md5_str
)


def test_get_seconds_from_epoch():
    assert get_seconds_from_epoch('Sat, 07 Mar 2020 21:03:07 GMT') == 1583614987.0
    assert get_seconds_from_epoch('Sat Mar  7 15:05:07 PST 2020') == 1583622307.0
    assert get_seconds_from_epoch('2017-07-01T14:59:55.711Z') == 1498921195.711
    assert get_seconds_from_epoch('2017-03-10 14:30:12,655+0000') == 1489156212.655


def test_base64_to_hex():
    assert base64_to_hex('aGVsbG8gd29ybGQ=') == hexlify('hello world'.encode()).decode()
    assert base64_to_hex('YXNkZnNkZnNhZGYjQEtKV0xLSkFTRDw+S0ZKQCNBU0RDQ0FTREBAIyRSIQ==') == hexlify('asdfsdfsadf#@KJWLKJASD<>KFJ@#ASDCCASD@@#$R!'.encode()).decode()
    assert base64_to_hex('NWEyNjUzYTY2Zjk1OGY5ZmNiODYwMGUyMDI4MTFiMjc=') == hexlify('5a2653a66f958f9fcb8600e202811b27'.encode()).decode()


def test_parse_md5_str():
    assert parse_md5_str('WiZTpm+Vj5/LhgDiAoEbJw==') == '5a2653a66f958f9fcb8600e202811b27'
    assert parse_md5_str('5a2653a66f958f9fcb8600e202811b27') == '5a2653a66f958f9fcb8600e202811b27'



#!/usr/bin/env python3
"""Race condition test will be done in test_race_cond.py
Here, we test 
"""
import os
import pytest
import time
from filelock import Timeout, BaseFileLock

from autouri.autouri import AutoURI, URIBase


class AutoURIFileLockTestException(Exception):
	pass


def test_autouri_lock(
	local_v6_txt,
	gcs_v6_txt,
	s3_v6_txt):
	for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):
		u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)
		lock = AutoURI(v6_txt).get_lock(no_lock=False)

		lock.acquire()
		try:
			assert u_lock.exists
			time.sleep(1)
		finally:
			lock.release()
		assert not u_lock.exists

		# trivial dummy lock
		lock = AutoURI(v6_txt).get_lock(no_lock=True)
		assert not isinstance(lock, BaseFileLock)


def test_autouri_lock_raise(
	local_v6_txt,
	gcs_v6_txt,
	s3_v6_txt):
	for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):		
		u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)
		lock = AutoURI(v6_txt).get_lock(no_lock=False)

		lock.acquire()
		try:
			assert u_lock.exists
			time.sleep(1)
			raise AutoURIFileLockTestException
		except AutoURIFileLockTestException:
			assert True
		else:
			assert False
		finally:
			lock.release()
			assert not u_lock.exists


def test_autouri_lock_with_context(
	local_v6_txt,
	gcs_v6_txt,
	s3_v6_txt):
	for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):		
		u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

		with AutoURI(v6_txt).get_lock(no_lock=False) as lock:
			assert u_lock.exists
			time.sleep(1)
		assert not u_lock.exists

		with AutoURI(v6_txt).get_lock(no_lock=True) as lock:
			assert not u_lock.exists
			time.sleep(1)
		assert not u_lock.exists


def test_autouri_lock_with_context_raise(
	local_v6_txt,
	gcs_v6_txt,
	s3_v6_txt):
	for v6_txt in (local_v6_txt, gcs_v6_txt, s3_v6_txt):		
		u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

		try:
			with AutoURI(v6_txt).get_lock(no_lock=False) as lock:
				assert u_lock.exists
				time.sleep(1)
				raise AutoURIFileLockTestException
		except AutoURIFileLockTestException:
			assert not u_lock.exists
		else:
			assert False


def test_autouri_lock_timeout(
	local_v6_txt):
	"""Timeout = 3, 8 sec
	For local storage (AbsPath) only.
	Default poll_interval (10 sec) is too long for test remote files.
	"""
	for v6_txt in (local_v6_txt,):
		u_lock = AutoURI(v6_txt + URIBase.LOCK_FILE_EXT)

		time_s = time.time()
		lock = AutoURI(v6_txt).get_lock(no_lock=False)
		lock.acquire()
		try:
			lock2 = AutoURI(v6_txt).get_lock(no_lock=False, timeout=3)
			lock2.acquire()
			try:
				pass
			finally:
				lock2.release()
		except Timeout:
			assert 2 < time.time() - time_s < 4
		else:
			assert False			
		finally:
			lock.release()
		assert not u_lock.exists

		time_s = time.time()
		lock = AutoURI(v6_txt).get_lock(no_lock=False)
		lock.acquire()
		try:
			lock2 = AutoURI(v6_txt).get_lock(no_lock=False, timeout=8)
			lock2.acquire()
			try:
				pass
			finally:
				lock2.release()
		except Timeout:
			assert 7 < time.time() - time_s < 9
		else:
			assert False			
		finally:
			lock.release()
		assert not u_lock.exists



import hashlib
import logging
import requests
from typing import Optional
from .autouri import URIBase, AutoURI, logger
from .metadata import URIMetadata, get_seconds_from_epoch, parse_md5_str


logger = logging.getLogger(__name__)


class ReadOnlyStorageError(Exception):
    pass


class HTTPURL(URIBase):
    """
    Class constants:
        HTTP_CHUNK_SIZE:
            Dict to replace path prefix with URL prefix.
            Useful to convert absolute path into URL on a web server.
    """
    HTTP_CHUNK_SIZE: int = 256*1024

    _LOC_SUFFIX = '.url'
    _SCHEMES = ('http://', 'https://')

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)

    @property
    def loc_dirname(self):
        """Dirname of URL is not very meaningful.
        Therefore, hash string of the whole URL string is used instead for localization.
        """
        return hashlib.md5(self._uri.encode('utf-8')).hexdigest()

    @property
    def basename(self):
        """Parses a URL to get a basename.
        This class can only work with a URL with an explicit basename
        which can be suffixed with extra parameters starting with ? only.
        """
        return super().basename.split('?', 1)[0]

    def _get_lock(self, timeout=None, poll_interval=None):
        raise ReadOnlyStorageError('Cannot lock on a read-only storage. {f}'.format(
            f=self._uri))

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        """Known issues about mtime:

        For URLs hosted on GCS (Google Cloud Storage) buckets, mtime will point to creation time
        if GCS object already existed and is modified. mtime of GCS object itself is accurate
        but corresponding URL on a public bucket will still have
        "Last-modified" property which is pointing to creation time.
        """
        ex, mt, sz, md5 = False, None, None, None
        try:
            # get header only
            r = requests.get(
                self._uri, stream=True, allow_redirects=True,
                headers=requests.utils.default_headers())
            r.raise_for_status()
            # make keys lower-case
            h = {k.lower(): v for k, v in r.headers.items()}
            ex = True

            if not skip_md5:
                if 'content-md5' in h:
                    md5 = parse_md5_str(h['content-md5'])
                elif 'x-goog-hash' in h:
                    hashes = h['x-goog-hash'].strip().split(',')
                    for hs in hashes:
                        if hs.strip().startswith('md5='):
                            raw = hs.strip().replace('md5=', '', 1)
                            md5 = parse_md5_str(raw)
                if md5 is None and 'etag' in h:
                    md5 = parse_md5_str(h['etag'])
                if md5 is None:
                    md5 = self.md5_from_file

            if 'content-length' in h:
                sz = int(h['content-length'])
            elif 'x-goog-stored-content-length' in h:
                sz = int(h['x-goog-stored-content-length'])

            if 'last-modified' in h:
                mt = get_seconds_from_epoch(h['last-modified'])

        except requests.exceptions.ConnectionError:
            pass
        except requests.exceptions.HTTPError as e:
            status_code = e.response.status_code
            if status_code == 403:
                raise

        return URIMetadata(
            exists=ex,
            mtime=mt,
            size=sz,
            md5=md5)

    def read(self, byte=False):
        r = requests.get(
            self._uri, stream=True, allow_redirects=True,
            headers=requests.utils.default_headers())
        r.raise_for_status()
        b = r.content
        if byte:
            return b
        else:
            return b.decode()

    def _write(self, s):
        raise ReadOnlyStorageError('Cannot write on a read-only storage. {f}'.format(
            f=self._uri))

    def _rm(self):
        raise ReadOnlyStorageError('Cannot remove a file on a read-only storage. {f}'.format(
            f=self._uri))

    def _cp(self, dest_uri):
        """Copy from HTTPURL to 
            AbsPath
        """
        from autouri.abspath import AbsPath
        dest_uri = AutoURI(dest_uri)

        if isinstance(dest_uri, AbsPath):
            r = requests.get(
                self._uri, stream=True, allow_redirects=True,
                headers=requests.utils.default_headers())
            r.raise_for_status()
            dest_uri.mkdir_dirname()
            with open(dest_uri._uri, 'wb') as f:
                for chunk in r.iter_content(chunk_size=HTTPURL.HTTP_CHUNK_SIZE): 
                    if chunk:
                        f.write(chunk)
            return True
        return False

    def _cp_from(self, src_uri):
        raise ReadOnlyStorageError('Cannot copy to a read-only storage. {f}'.format(
            f=self._uri))

    @staticmethod
    def get_http_chunk_size() -> int:
        return HTTPURL.HTTP_CHUNK_SIZE

    @staticmethod
    def init_httpurl(
        http_chunk_size: Optional[int]=None):
        if http_chunk_size is not None:
            HTTPURL.HTTP_CHUNK_SIZE = http_chunk_size
        if HTTPURL.HTTP_CHUNK_SIZE % (256*1024) > 0:
            raise ValueError(
                'HTTPURL.HTTP_CHUNK_SIZE must be a multiple of 256 KB (256*1024) '
                'to be compatible with cloud storage APIs (GCS and AWS S3).')





"""S3 Bucket configuration:
    S3 Object versioning must be turned off
"""
import logging
import requests
import time
from boto3 import client
from botocore.exceptions import ClientError
from filelock import BaseFileLock
from tempfile import NamedTemporaryFile
from typing import Tuple, Optional
from .autouri import URIBase, AutoURI
from .metadata import URIMetadata, get_seconds_from_epoch, parse_md5_str


logger = logging.getLogger(__name__)


class S3URILock(BaseFileLock):
    """Locking without using S3 Object Lock.
    Without S3 object lock, boto3's put_object(), which is used for _write() and _cp() in this module,
    does not ensure consistency of multiple write operations at the same time.
    It overwrites for all write requests but the last object written.
    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Client.put_object

    To make this lock as stable as possible, this module uses .lock file with id(self) written on it.
    This module first checks if .lock does not exist, then tries to write .lock with id(self).
    It waits for a short time (self._lock_read_delay) and checks if written .lock has the same id(self).
    self._lock_read_delay is set as poll_interval/10.
    """
    def __init__(
        self, lock_file, timeout=900, poll_interval=10.0, no_lock=False):
        super().__init__(lock_file, timeout=timeout)
        self._poll_interval = poll_interval
        self._lock_read_delay = self._poll_interval/10.0

    def acquire(self, timeout=None, poll_intervall=5.0):
        """To use self._poll_interval instead of poll_intervall in args.
        """
        super().acquire(timeout=timeout, poll_intervall=self._poll_interval)

    def _acquire(self):
        """Unlike GCSURI, this module does not use S3 Object locking.
        This will write id(self) on a .lock file.
        """
        u = S3URI(self._lock_file)
        str_id = str(id(self))
        try:
            if not u.exists:
                u.write(str_id, no_lock=True)
                time.sleep(self._lock_read_delay)
            if u.read() == str_id:
                self._lock_file_fd = id(self)
        except ClientError as e:
            status = e.response["ResponseMetadata"]["HTTPStatusCode"]
            if status in (403, 404):
                raise
        return None

    def _release(self):
        u = S3URI(self._lock_file)
        try:
            u.rm(no_lock=True)
            self._lock_file_fd = None
        except ClientError:
            pass
        return None


class S3URI(URIBase):
    """
    Class constants:
        LOC_PREFIX (inherited):
            Path prefix for localization. Inherited from URIBase class.
        DURATION_PRESIGNED_URL:
            Duration for presigned URLs in seconds.

    Protected class constants:
        _CACHED_BOTO3_CLIENT_PER_THREAD:
        _CACHED_PRESIGNED_URLS:
        _S3_PUBLIC_URL_FORMAT:
            End point for a bucket with public access + key path
    """
    DURATION_PRESIGNED_URL: int = 4233600

    _CACHED_BOTO3_CLIENT_PER_THREAD = {}
    _CACHED_PRESIGNED_URLS = {}
    _S3_PUBLIC_URL_FORMAT = 'http://{bucket}.s3.amazonaws.com/{path}'

    _LOC_SUFFIX = '.s3'
    _SCHEMES = ('s3://',)

    def __init__(self, uri, thread_id=-1):
        super().__init__(uri, thread_id=thread_id)

    def _get_lock(self, timeout=None, poll_interval=None):
        if timeout is None:
            timeout = S3URI.LOCK_TIMEOUT
        if poll_interval is None:
            poll_interval = S3URI.LOCK_POLL_INTERVAL
        return S3URILock(
            self._uri + S3URI.LOCK_FILE_EXT,
            timeout=timeout,
            poll_interval=poll_interval)

    def get_metadata(self, skip_md5=False, make_md5_file=False):
        ex, mt, sz, md5 = False, None, None, None

        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        try:
            m = cl.head_object(Bucket=bucket, Key=path)['ResponseMetadata']['HTTPHeaders']
            # make keys lower-case
            h = {k.lower(): v for k, v in m.items()}
            ex = True

            if not skip_md5:
                if 'content-md5' in h:
                    md5 = parse_md5_str(h['content-md5'])
                elif 'etag' in h:
                    md5 = parse_md5_str(h['etag'])
                if md5 is None:
                    # make_md5_file is ignored for S3URI
                    md5 = self.md5_from_file                    

            if 'content-length' in h:
                sz = int(h['content-length'])

            if 'last-modified' in h:
                mt = get_seconds_from_epoch(h['last-modified'])

        except Exception as e:
            pass

        return URIMetadata(
            exists=ex,
            mtime=mt,
            size=sz,
            md5=md5)

    def read(self, byte=False):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        obj = cl.get_object(Bucket=bucket, Key=path)
        if byte:
            return obj['Body'].read()
        return obj['Body'].read().decode()

    def _write(self, s):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(s, str):
            b = s.encode('ascii')
        else:
            b = s
        cl.put_object(Bucket=bucket, Key=path, Body=b)
        return

    def _rm(self):
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        cl.delete_object(Bucket=bucket, Key=path)
        return

    def _cp(self, dest_uri):
        """Copy from S3URI to 
            S3URI
            AbsPath
        """
        from .abspath import AbsPath

        dest_uri = AutoURI(dest_uri)
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(dest_uri, S3URI):
            dest_bucket, dest_path = dest_uri.get_bucket_path()
            cl.copy_object(
                CopySource={
                    'Bucket': bucket,
                    'Key': path},
                Bucket=dest_bucket,
                Key=dest_path)
            return True

        elif isinstance(dest_uri, AbsPath):
            dest_uri.mkdir_dirname()
            with open(dest_uri._uri, 'wb') as fp:
                cl.download_fileobj(Bucket=bucket, Key=path, Fileobj=fp)
            return True
        return False

    def _cp_from(self, src_uri):
        """Copy to S3URI from
            AbsPath
            HTTPURL
        """
        from .abspath import AbsPath
        from .httpurl import HTTPURL

        src_uri = AutoURI(src_uri)
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()

        if isinstance(src_uri, AbsPath):
            p = cl.upload_file(
                Filename=src_uri._uri,
                Bucket=bucket,
                Key=path)
            return True

        elif isinstance(src_uri, HTTPURL):
            r = requests.get(
                src_uri._uri, stream=True, allow_redirects=True,
                headers=requests.utils.default_headers())
            r.raise_for_status()
            with NamedTemporaryFile() as fp:
                for chunk in r.iter_content(HTTPURL.get_http_chunk_size()):
                    fp.write(chunk)
                fp.seek(0)
                cl.upload_fileobj(
                    Fileobj=fp,
                    Bucket=bucket,
                    Key=path)
            return True        
        return False

    def get_bucket_path(self) -> Tuple[str, str]:
        """Returns a tuple of URI's S3 bucket and path.
        """
        bucket, path = self.uri_wo_scheme.split(S3URI.get_path_sep(), 1)
        return bucket, path

    def get_presigned_url(self, duration=None, use_cached=False) -> str:
        """
        Args:
            duration: Duration in seconds. This is ignored if use_cached is on.
            use_cached: Use a cached URL. 
        """
        cache = S3URI._CACHED_PRESIGNED_URLS
        if use_cached:
            if cache is not None and self._uri in cache:
                return cache[self._uri]
        cl = S3URI.get_boto3_client(self._thread_id)
        bucket, path = self.get_bucket_path()
        duration = duration if duration is not None else S3URI.DURATION_PRESIGNED_URL
        url = cl.generate_presigned_url(
            'get_object',
            Params={'Bucket': bucket, 'Key': path},
            ExpiresIn=duration)
        cache[self._uri] = url
        return url

    def get_public_url(self) -> str:
        bucket, path = self.get_bucket_path()
        return S3URI._S3_PUBLIC_URL_FORMAT.format(bucket=bucket, path=path)

    @staticmethod
    def get_boto3_client(thread_id=-1) -> client:
        if thread_id in S3URI._CACHED_BOTO3_CLIENT_PER_THREAD:
            return S3URI._CACHED_BOTO3_CLIENT_PER_THREAD[thread_id]
        else:
            cl = client('s3')
            S3URI._CACHED_BOTO3_CLIENT_PER_THREAD[thread_id] = cl
            return cl

    @staticmethod
    def init_s3uri(
        loc_prefix: Optional[str]=None,
        duration_presigned_url: Optional[int]=None):
        if loc_prefix is not None:
            S3URI.LOC_PREFIX = loc_prefix
        if duration_presigned_url is not None:
            S3URI.DURATION_PRESIGNED_URL = duration_presigned_url




import setuptools


with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='autouri',
    version='0.1.4',
    python_requires='>=3.6',
    scripts=['bin/autouri'],
    author='Jin wook Lee',
    author_email='leepc12@gmail.com',
    description='Automatic localization for various URIs (s3://, gs://, http://, https:// and local path)',
    long_description='https://github.com/ENCODE-DCC/autouri',
    long_description_content_type='text/markdown',
    url='https://github.com/ENCODE-DCC/autouri',
    packages=setuptools.find_packages(exclude=['docs', 'test']),
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
    ],
    install_requires=[
        'requests', 'pyopenssl',
        'google-cloud-storage',
        'boto3', 'awscli',
        'dateparser', 'filelock'
    ]
)



#!/usr/bin/env python3
import argparse
import logging
import os
import sys
from .autouri import URIBase, AutoURI
from .abspath import AbsPath
from .httpurl import HTTPURL
from .s3uri import S3URI
from .gcsuri import GCSURI
from . import __version__ as version


logger = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        '-v', '--version', action='store_true',
        help='Show version')

    parent_all = argparse.ArgumentParser(add_help=False)
    group_log_level = parent_all.add_mutually_exclusive_group()
    group_log_level.add_argument('-D', '--debug', action='store_true',
                   help='Prints all logs >= DEBUG level')

    parent_src = argparse.ArgumentParser(add_help=False)
    parent_src.add_argument('src',
        help='Source file URI')

    parent_target = argparse.ArgumentParser(add_help=False)
    parent_target.add_argument('target',
        help='Target file/directory URI (e.g. gs://here/me.txt, s3://here/my-dir/) .'
             'Directory must have a trailing directory separator '
             '(e.g. /hello/. gs://where/am/i/).')

    parent_cp = argparse.ArgumentParser(add_help=False)
    parent_cp.add_argument('--use-gsutil-for-s3', action='store_true',
        help='Use gsutil for DIRECT TRANSFER between gs:// and s3://. '
             'gsutil must be installed and configured to have AWS credentials '
             'in ~/.boto file. Run "gsutil config" do generate it.')
    parent_cp.add_argument('--make-md5-file', action='store_true',
        help='Make .md5 file to store file\'s md5 hexadecimal string. '
             'This file can be used later to prevent repeated md5sum calculation. '
             'This is for local path only since cloud URIs already provide md5 hash '
             'info in HTTP headers.')

    subparser = parser.add_subparsers(dest='action')

    p_metadata = subparser.add_parser(
        'metadata',
        help='AutoURI(src).get_metadata(): Get metadata of source.',
        parents=[parent_src, parent_all])

    p_cp = subparser.add_parser(
        'cp',
        help='AutoURI(src).cp(target): Copy source to target. '
             'target must be a full filename/directory. '
             'Target directory must have a trailing directory separator '
             '(e.g. /)',
        parents=[parent_src, parent_target, parent_cp, parent_all])

    p_read = subparser.add_parser(
        'read',
        help='AutoURI(src).read(): Read from source.',
        parents=[parent_src, parent_all])

    p_write = subparser.add_parser(
        'write',
        help='AutoURI(src).write(text): Write text on source.',
        parents=[parent_src, parent_all])
    p_write.add_argument('text',
        help='Text to be written to source file.')

    p_rm = subparser.add_parser(
        'rm',
        help='AutoURI(src).rm(): Delete source.',
        parents=[parent_src, parent_all])

    p_loc = subparser.add_parser(
        'loc',
        help='AutoURI(src).localize_on(target): Localize source on target directory '
             'Target directory must end with directory separator',
        parents=[parent_src, parent_target, parent_cp, parent_all])
    p_loc.add_argument('--recursive', action='store_true',
        help='Recursively localize source into target directory.')

    p_presign = subparser.add_parser(
        'presign',
        help='AutoURI(src).get_presigned_url(). For cloud-based URIs only.',
        parents=[parent_src, parent_all])
    p_presign.add_argument('--gcp-private-key-file',
        help='GCP private key file (JSON format) to presign gs:// URIs.')
    p_presign.add_argument('--duration', type=int,
        help='Duration of presigned URL in seconds.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()
    if args.version:
        print(version)
        parser.exit()

    if args.debug:
        log_level = 'DEBUG'
    else:
        log_level = 'INFO'
    logging.basicConfig(
        level=log_level,
        format='%(asctime)s|%(name)s|%(levelname)s| %(message)s')
    # suppress filelock logging
    logging.getLogger('filelock').setLevel('CRITICAL')

    return args


def get_local_path_if_valid(s):
    abspath = os.path.abspath(os.path.expanduser(s))
    dirname = os.path.dirname(abspath)
    if os.path.isdir(abspath):
        return abspath + os.sep
    elif os.path.exists(abspath):
        return abspath
    elif os.path.exists(dirname):
        return abspath
    return s


def main():
    args = parse_args()

    src = get_local_path_if_valid(args.src)

    if args.action in ('cp', 'loc'):
        if args.use_gsutil_for_s3:
            GCSURI.init_gcsuri(use_gsutil_for_s3=True)
        target = get_local_path_if_valid(args.target)

    if args.action == 'metadata':
        m = AutoURI(src).get_metadata()
        print(m)

    elif args.action == 'cp':
        u_src = AutoURI(src)
        _, flag = u_src.cp(
            target, make_md5_file=args.make_md5_file, return_flag=True)

        if flag == 0:
            logger.info('Copying from file {s} to {t} done'.format(
                s=src, t=target))
        elif flag:
            if flag == 1:
                reason = 'skipped due to md5 hash match'
            elif flag == 2:
                reason = 'skipped due to filename/size match and mtime test'
            else:
                raise NotImplementedError
            logger.info('Copying from file {s} to {t} {reason}'.format(
                s=src, t=target, reason=reason))

    elif args.action == 'read':
        s = AutoURI(src).read()
        print(s)

    elif args.action == 'write':
        AutoURI(src).write(args.text)
        logger.info('Text has been written to {s}'.format(s=src))

    elif args.action == 'rm':        
        u = AutoURI(src)
        if not u.exists:
            raise ValueError('File does not exist. {s}'.format(s=src))
        u.rm()
        logger.info('Deleted {s}'.format(s=src))

    elif args.action == 'loc':
        _, localized = AutoURI(src).localize_on(
            target,
            recursive=args.recursive,
            make_md5_file=args.make_md5_file,
            return_flag=True)
        if localized:
            logger.info('Localized {s} on {t}'.format(s=src, t=target))
        else:
            logger.info('No need to localize {s} on {t}'.format(s=src, t=target))

    elif args.action == 'presign':
        u = AutoURI(src)

        if isinstance(u, GCSURI):
            if not args.gcp_private_key_file:
                raise ValueError('GCP private key file (--gcp-private-key-file) not found.')
            url = u.get_presigned_url(duration=args.duration,
                                      private_key_file=args.gcp_private_key_file)
            print(url)

        elif isinstance(u, S3URI):
            url = u.get_presigned_url(duration=args.duration)
            print(url)

        else:
            raise ValueError('Presigning URL is available for cloud URIs (gs://, s3://) only.')


if __name__ == '__main__':
    main()
